# Django
from django.db.models import QuerySet
from django.core.exceptions import ImproperlyConfigured
from django.utils.text import slugify
from django.urls import path

# Views
from .views import (
    ListView,
    CreateView,
    UpdateView,
    MassUpdateView,
    DetailView,
    DeleteView,
    MassDeleteView,
    DuplicateView,
)

from . import settings

ALL_FIELDS = "__all__"


class ModelSite:
    """Superclass that generate CRUD Views for any model"""

    # Views
    model = None
    form_class = None  # Used for create Create and Update views
    fields = None  # User for passed to Create and Update views for generate forms
    list_fields = ("__str__",)  # Used for create ListView with de specified fields
    detail_fields = ()  # Used for create DetailView with specified fields
    allow_views = (
        "list",
        "create",
        "update",
        "detail",
        "delete",
    )  # Says superadmin which views create
    create_success_url = "list"
    update_success_url = "list"
    delete_success_url = "list"

    # Context
    list_extra_context = {}
    form_extra_context = {}
    detail_extra_context = {}

    # Inlines
    inlines = {}

    # Templates
    list_template_name = None  # Says which list template use
    form_template_name = None  # Says which form template use
    detail_template_name = None  # Says which detail template use
    delete_template_name = None  # Says which delete template use

    # Mixins
    list_mixins = ()  # List of mixins that include in ListViews
    form_mixins = ()  # List of mixins that include in Create and Update Views
    detail_mixins = ()  # List of mixins that include in DetailViews
    delete_mixins = ()  # List of mixins that include in DetailViews

    # Prepopulate
    slug_field = "slug"
    prepopulate_slug = ()

    # Options for build queryset
    queryset = None  # Specified custom queryset
    paginate_by = None  # Specified if ListView paginated by

    # Filter and ordering
    search_fields = ()  # Used for create search method by specified fields
    filter_fields = ()
    order_by = ()  # User for crate ordering methods by specified fields

    # Urls
    url_list_suffix = "list"
    url_create_suffix = "create"
    url_update_suffix = "update"
    url_detail_suffix = "detail"
    url_delete_suffix = "delete"
    url_duplicate_suffix = "duplicate"

    url_mass_update_suffix = "mass_update"
    url_mass_delete_suffix = "mass_delete"

    def __init__(self, model, **kwargs):
        self.model = model
        if not self.model:
            raise ImproperlyConfigured("The 'model' attribute must be specified.")

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not isinstance(self.queryset, QuerySet):
            self.queryset = self.model._default_manager.all()

        if not isinstance(self.allow_views, tuple):
            raise ImproperlyConfigured("The 'allow_views' attribute must be a tuple.")

        if not self.form_class and not self.fields:
            self.fields = ALL_FIELDS

        self.breadcrumb_home_text = getattr(
            self, "breadcrumb_home_text", settings.BREADCRUMB_HOME_TEXT
        )
        self.breadcrumb_create_text = getattr(
            self, "breadcrumb_create_text", settings.BREADCRUMB_CREATE_TEXT
        )
        self.breadcrumb_update_text = getattr(
            self, "breadcrumb_update_text", settings.BREADCRUMB_UPDATE_TEXT
        )
        self.breadcrumb_detail_text = getattr(
            self, "breadcrumb_detail_text", settings.BREADCRUMB_DETAIL_TEXT
        )
        self.breadcrumb_delete_text = getattr(
            self, "breadcrumb_delete_text", settings.BREADCRUMB_DELETE_TEXT
        )

    def get_info(self):
        """Obtiene la información del modelo"""
        info = self.model._meta.app_label, self.model._meta.model_name
        # info = slugify(self.model._meta.app_config.verbose_name), slugify(
        #    self.model._meta.verbose_name
        # )
        return info

    # Url methods
    def get_base_url_name(self, suffix):
        info = self.get_info()
        url_suffix = getattr(self, "url_%s_suffix" % suffix)
        base_url_name = "%s_%s_%s" % (*info, url_suffix)
        return base_url_name

    def get_url_name(self, suffix):
        url_name = "site:%s" % self.get_base_url_name(suffix)
        return url_name

    def get_urls(self):
        """Genera las urls para los modelos registrados"""

        # def wrap(view):
        #     def wrapper(*args, **kwargs):
        #         return self.admin_site.admin_view(view)(*args, **kwargs)
        #     wrapper.model_admin = self
        #     return update_wrapper(wrapper, view)
        urlpatterns = []

        has_slug = hasattr(self.model, self.slug_field)
        route_param = "<slug:slug>" if has_slug else "<int:pk>"

        if "list" in self.allow_views:
            # url_name = "%s_%s_%s" % (*info, self.url_list_suffix)
            url_name = self.get_base_url_name("list")
            urlpatterns += [
                path(route="", view=ListView.as_view(site=self), name=url_name)
            ]

        if "create" in self.allow_views:
            url_create_name = self.get_base_url_name("create")

            urlpatterns += [
                path(
                    route=f"{self.url_create_suffix}/",
                    view=CreateView.as_view(site=self),
                    name=url_create_name,
                ),
            ]

        if "update" in self.allow_views:
            url_update_name = self.get_base_url_name("update")

            urlpatterns += [
                path(
                    route=f"{route_param}/{self.url_update_suffix}/",
                    view=UpdateView.as_view(site=self),
                    name=url_update_name,
                ),
                path(
                    route=f"{self.url_update_suffix}/",
                    view=MassUpdateView.as_view(site=self),
                    name=self.get_base_url_name("mass_update"),
                ),
            ]

        if "detail" in self.allow_views:
            url_detail_suffix = self.url_detail_suffix
            url_detail = f"{route_param}/"
            if url_detail_suffix:
                url_detail = f"{url_detail}/{self.url_detail_suffix}/"
            url_detail_name = self.get_base_url_name("detail")
            urlpatterns += [
                path(
                    route=url_detail,
                    view=DetailView.as_view(site=self),
                    name=url_detail_name,
                ),
            ]

        if "delete" in self.allow_views:
            url_delete_name = self.get_base_url_name("delete")

            urlpatterns += [
                path(
                    route=f"{route_param}/{self.url_delete_suffix}/",
                    view=DeleteView.as_view(site=self),
                    name=url_delete_name,
                ),
                path(
                    route=f"{self.url_delete_suffix}/",
                    view=MassDeleteView.as_view(site=self),
                    name=self.get_base_url_name("mass_delete"),
                ),
            ]

        urlpatterns += [
            path(
                route=f"{route_param}/{self.url_duplicate_suffix}/",
                view=DuplicateView.as_view(site=self),
                name=self.get_base_url_name("duplicate"),
            ),
        ]

        # urlpatterns = [
        # path('add/', wrap(self.add_view), name='%s_%s_add' % info),
        # path('autocomplete/', wrap(self.autocomplete_view), name='%s_%s_autocomplete' % info),
        # path('<path:object_id>/history/', wrap(self.history_view), name='%s_%s_history' % info),
        # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
        # path('<path:object_id>/change/', wrap(self.change_view), name='%s_%s_change' % info),
        # # For backwards compatibility (was the change url before 1.9)
        # path('<path:object_id>/', wrap(RedirectView.as_view(
        #     pattern_name='%s:%s_%s_change' % ((self.admin_site.name,) + info)
        # ))),
        # ]
        return urlpatterns

    @property
    def urls(self):
        """Return created urls"""
        return self.get_urls()



""" """
# Django
from django.views.generic import CreateView as BaseCreateView
from django.forms.models import model_to_dict

# Local
from .base import SiteView, get_base_view
from ..shortcuts import get_object_from_site, get_urls_of_site
from ..utils import import_all_mixins, import_mixin


class CreateMixin:
    """Definimos la clase que utilizará el modelo"""

    # permission_required = permission_autosite + self.permission_extra

    action = "create"
    duplicate_param = "duplicate"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.site.form_extra_context)
        return context

    def get_related_initial(self, object):
        related_initial = {}
        for related in object._meta.related_objects:
            related_name = related.related_name
            related_name = related_name if related_name else f"{related.name}_set"
            related_objects = [
                model_to_dict(
                    obj,
                    fields=[
                        field.name
                        for field in obj._meta.fields
                        if field.name != "id"
                        and field.name != related.remote_field.name
                    ],
                )
                for obj in getattr(object, related_name).all()
            ]
            related_initial.update({related.related_model: related_objects})

        return related_initial

    def get_initial(self):
        initial = super().get_initial()
        if self.request.method == "GET":
            slug_or_pk = self.request.GET.get(self.duplicate_param)
            if slug_or_pk:
                object = get_object_from_site(self.site, slug_or_pk)
                if object:
                    data = model_to_dict(
                        object,
                        fields=[
                            field.name
                            for field in object._meta.fields
                            if field.name != "id"
                        ],
                    )
                    initial.update(data)
                    initial["related_initial"] = self.get_related_initial(object)
        return initial

    def get_success_url(self):
        urls = get_urls_of_site(self.site, object=self.object)
        return urls.get(self.site.create_success_url)


class CreateView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        mixins = import_all_mixins() + [CreateMixin]
        if self.site.inlines and isinstance(self.site.inlines, (list, tuple, dict)):
            InlinesMixin = import_mixin("InlinesMixin")

            class Inlines(InlinesMixin):
                inlines = self.site.inlines

            mixins += [Inlines]
        View = get_base_view(BaseCreateView, mixins, self.site)

        # Set attributes
        View.form_class = self.site.form_class
        View.fields = self.site.fields

        View.__bases__ = (*self.site.form_mixins, *View.__bases__)
        view = View.as_view()
        return view(request, *args, **kwargs)



""" """
# Django
from django.views.generic import RedirectView
from django.urls import reverse


class DuplicateView(RedirectView):
    site = None
    duplicate_param = "duplicate"

    def get_redirect_url(self, *args, **kwargs):
        slug_or_pk = kwargs.get("slug") if "slug" in kwargs else kwargs.get("pk")
        url = reverse(self.site.get_url_name("create"))
        url = "%s?%s" % (url, f"{self.duplicate_param}={slug_or_pk}")
        return url



class TemplateMixin:
    def get_template_names(self):
        action = self.template_name_suffix.split("_")[-1]
        self.template_name = getattr(self.site, "%s_template_name" % action)
        return super().get_template_names()



"""  Module Signals """

# Django
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import pre_save, post_save
from django.dispatch import receiver
from django.utils.text import slugify
from django.apps import apps

# Models
from .models import Menu

# Local
from .services import FieldService
from . import site


""" Signal for presave instance """


@receiver(post_save)
def prepopulate_slug(sender, instance, **kwargs):
    if not site.is_registered(sender):
        return

    model_site = site.get_modelsite(sender)
    model_name = sender._meta.model_name

    slug_fields = model_site.prepopulate_slug

    if slug_fields and not isinstance(slug_fields, tuple):
        raise ImproperlyConfigured("Field 'prepopulate_slug' must be a tuple")

    if not slug_fields:
        return

    if not hasattr(sender, "slug"):
        raise ImproperlyConfigured(f"Model '{model_name}' has not 'slug' field")

    for field in slug_fields:
        if not hasattr(sender, field):
            raise ImproperlyConfigured(
                f"Model '{model_name}' has no field'{str(field)}'"
            )

    fields = (FieldService.get_field_value(instance, field) for field in slug_fields)
    slug = slugify(" ".join(fields))
    if instance.slug != slug:
        instance.slug = slug
        instance.save()


@receiver(pre_save, sender=Menu)
def add_route(sender, instance, **kwargs):
    instance.route = instance.get_route()


@receiver(post_save, sender=Menu)
def check(sender, instance, **kwargs):

    post_save.disconnect(check, sender=Menu)
    pre_save.disconnect(add_route, sender=Menu)

    def update_route(menu):
        menu.route = menu.get_route()
        menu.save()
        for submenu in menu.submenus.all():
            update_route(submenu)

    def update_groups():
        for menu in Menu.objects.all():
            menu.is_group = bool(menu.submenus.count())
            menu.save()

    update_route(instance)
    update_groups()

    post_save.connect(check, sender=Menu)
    pre_save.connect(add_route, sender=Menu)



""" List view engine"""
# Django
from django.views.generic import ListView as BaseListView


# Local
from .base import SiteView, get_base_view
from ..shortcuts import get_urls_of_site
from ..utils import import_mixin, import_all_mixins

# Utilities
from ..services import FieldService


class ListMixin:
    """Definimos la clase que utilizará el modelo"""

    action = "list"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.site.list_extra_context)

        opts = {
            "fields": self.get_list_fields(),
            "rows": self.get_rows(context["object_list"]),
            "page_start_index": context["page_obj"].start_index()
            if context["is_paginated"]
            else 1,
            "page_end_index": context["page_obj"].end_index()
            if context["is_paginated"]
            else context["object_list"].count(),
            "total_records": context["paginator"].count
            if context["is_paginated"]
            else context["object_list"].count(),
        }

        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})

        return context

    def get_paginate_by(self, queryset):
        paginate_by = self.request.GET.get("paginate_by")
        if paginate_by:
            return paginate_by
        return super().get_paginate_by(queryset)

    def get_list_fields(self):
        fields = [
            (name, FieldService.get_field_label(self.model, name))
            for name in self.site.list_fields
        ]
        return fields

    def get_editable_fields(self):
        fields = [
            (name, FieldService.get_field_label(self.model, name))
            for name in self.site.form_class._meta.fields
        ]
        return fields

    def get_rows(self, queryset):
        rows = [
            {
                "instance": instance,
                "values": self.get_values(instance),
                "urls": get_urls_of_site(
                    self.site, object=instance, user=self.request.user
                ),
            }
            for instance in queryset
        ]
        return rows

    def get_values(self, instance):
        values = [
            FieldService.get_field_value(instance, name)
            for name in self.site.list_fields
        ]
        return values


class ListView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        FilterMixin = import_mixin("FilterMixin")
        mixins = import_all_mixins() + [FilterMixin, ListMixin]
        View = get_base_view(BaseListView, mixins, self.get_site())

        # Set attriburtes
        View.paginate_by = self.site.paginate_by

        View.__bases__ = (*self.site.list_mixins, *View.__bases__)

        view = View.as_view()
        return view(request, *args, **kwargs)



# Django
from django.core.management.base import BaseCommand, CommandError

# local
from .base import create_actions


class Command(BaseCommand):
    help = "Create actions mapping all apps"

    def handle(self, *args, **options):
        create_actions()
        self.stdout.write(self.style.SUCCESS("Successfully actions was created"))



# Models
from .models import Menu
from .utils import get_user_menu


def menu(request):
    return {"menu_tree": build_user_menu(request.user)}


def build_user_menu(user):
    menu_list = list()
    if user.is_authenticated and user.is_active:
        object_list = Menu.objects.filter(parent__isnull=True)
        menu_list = get_user_menu(object_list, user)
    return menu_list



# Django
from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command

# Models
from superadmin.models import Action, Menu

#
from superadmin import site


class Command(BaseCommand):
    help = "Create base menu mapping all apps"

    def handle(self, *args, **options):
        call_command("createactions")

        Menu.objects.all().delete()

        default_action = Action.objects.get(
            app_label="superadmin", element="ModuleView"
        )

        apps = {}
        for model in site._registry:
            if model._meta.app_config in apps:
                apps[model._meta.app_config].append(model)
            else:
                apps[model._meta.app_config] = [model]

        sequence = 1
        for app in apps:
            menu = Menu.objects.create(
                name=app.verbose_name.capitalize(),
                action=default_action,
                is_group=True,
                sequence=sequence,
            )
            sequence += 1

            index = 1
            for model in apps[app]:
                action = Action.objects.get(
                    app_label=app.label, element=model._meta.model_name
                )

                Menu.objects.create(
                    parent=menu,
                    name=model._meta.verbose_name_plural.capitalize(),
                    action=action,
                    is_group=False,
                    sequence=index,
                )

                index += 1

        self.stdout.write(self.style.SUCCESS("Successfully base menu was created"))



""" """
# Python
from functools import reduce

# Django
from django.core.exceptions import ImproperlyConfigured
from django.views.generic import DetailView as BaseDetailView
from django.contrib.admin.utils import flatten

# Local
from .base import SiteView, get_base_view
from ..services import FieldService, FilterService
from ..utils import import_all_mixins
from ..shortcuts import get_urls_of_site


class DetailMixin:
    """Definimos la clase que utilizará el modelo"""

    action = "detail"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.site.detail_extra_context)
        flatten_results, fieldset_results = self.get_results()
        params = FilterService.get_params(self.site.model, self.request.session)
        queryset = FilterService.filter(self.site.queryset, params)
        nav = FilterService.get_previous_and_next(queryset, self.object)
        if nav:
            nav["previous_url"] = (
                get_urls_of_site(self.site, nav["previous"])[self.action]
                if nav.get("previous")
                else None
            )
            nav["next_url"] = (
                get_urls_of_site(self.site, nav["next"])[self.action]
                if nav.get("next")
                else None
            )
        opts = {
            "results": fieldset_results,
            "flatten_results": flatten_results,
            "nav": nav,
        }

        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})

        return context

    def get_results(self):
        if isinstance(self.site.detail_fields, (list, tuple)):
            fields = flatten(self.site.detail_fields)
        elif isinstance(self.site.detail_fields, dict):
            fields = reduce(
                lambda acc, fieldset: acc + flatten(fieldset),
                self.site.detail_fields.values(),
                [],
            )
        else:
            raise ImproperlyConfigured(
                "The fieldsets must be an instance of list, tuple or dict"
            )
        fields = fields if fields else (field.name for field in self.model._meta.fields)
        results = {
            field: (
                FieldService.get_field_label(self.object, field),
                FieldService.get_field_value(self.object, field),
                FieldService.get_field_type(self.object, field),
                field,
            )
            for field in fields
        }

        flatten_results = results.values()

        def parse(fieldset):
            def wrap(fields):
                fields = fields if isinstance(fields, (list, tuple)) else [fields]
                return {
                    "bs_cols": int(12 / len(fields)),
                    "fields": [results.get(field, ("", "", "")) for field in fields],
                }

            fieldset_list = list(map(wrap, fieldset))
            return fieldset_list

        fieldsets_list = self.site.detail_fields
        fieldsets = (
            [(None, fieldsets_list)]
            if isinstance(fieldsets_list, (list, tuple))
            else fieldsets_list.items()
        )
        fieldsets_results = [
            {"title": title or "", "fieldset": parse(fieldset)}
            for title, fieldset in fieldsets
        ]

        return flatten_results, fieldsets_results

    def get_slug_field(self):
        return self.site.slug_field or super().get_slug_field()


class DetailView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        mixins = import_all_mixins() + [DetailMixin]
        View = get_base_view(BaseDetailView, mixins, self.site)

        # Set attributes
        View.__bases__ = (*self.site.detail_mixins, *View.__bases__)

        view = View.as_view()
        return view(request, *args, **kwargs)
